//
// GeometricUtilities.java for vgi in /home/loulou/lrde_proj/vgi/src
//
// Made by Louis-Noel Pouchet
// Login   <loulou@lrde.epita.fr>
//
// Started on  Wed Mar  2 17:28:31 2005 Louis-Noel Pouchet
// Last update Fri Dec  8 02:28:38 2006 Guillaume
//

package vaucanson.vgitoolkit;

import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.CellView;
import org.jgraph.graph.EdgeView;
import org.jgraph.graph.PortView;
import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.GraphLayoutCache;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import java.util.Vector;
import java.util.Set;
import java.util.Iterator;
import java.util.List;


// Geometric functions.
//
public class GeometricUtilities
{
    public static final double		ANGLE_TOLERANCE = 2;
    public static final double		LOOP_ANGLE_TOLERANCE = 2;


    // Get free cardinal position, following preferred order.
    public static int			get_free_cardinal(VJGraph graph,
							  State s,
							  DefaultEdge edge)
    {
      State[] states = VJGraph.get_connected_states(graph, s);
      Transition[] loops = VJGraph.get_loops(graph, s);
      GraphLayoutCache cache = (GraphLayoutCache)graph.getGraphLayoutCache();
      Rectangle2D bounds = cache.getMapping(new Object[]{s})[0].getBounds();
      Point2D source = new Point2D.Double(bounds.getCenterX(),
					  bounds.getCenterY());
      int mask = 0;
      int i;
      int alpha;

      // Check if any connected state is at cardinal point.
      for (i = 0; i < states.length; ++i)
      {
	Rectangle2D b =
	  cache.getMapping(new Object[]{states[i]})[0].getBounds();
	Point2D p = new Point2D.Double(b.getCenterX(), b.getCenterY());
	alpha = compute_angle(source, p);
	mask = compute_cardinal_mask(mask, alpha, ANGLE_TOLERANCE);
      }

      // Check if any loop is at cardinal point.
      for (i = 0; i < loops.length; ++i)
      {
	if (loops[i] == null || loops[i] == edge)
	  continue;
	CellView[] cc = cache.getMapping(new Object[]{loops[i]});
	CellView c = cc[0];

	// Ensure mapping is present, else delay computation to next
	// call.
	if (c != null)
	{
	  Rectangle2D b = c.getBounds();
	  Point2D p = new Point2D.Double(b.getCenterX(),
					 b.getCenterY());
	  alpha = compute_angle(source, p);
	  mask = compute_cardinal_mask(mask, alpha,
				       LOOP_ANGLE_TOLERANCE);
	}
      }

      // Check if any initial or final transition is at cardinal point.
      if (s.is_initial() && edge != s.initial_transition_get())
      {
	set_state_initial_angle(graph, s);
	mask = compute_cardinal_mask(mask,
				     GeometryProperties.getTransitionAngle
				     (s.initial_transition_get()),
				     ANGLE_TOLERANCE);
      }
      if (s.is_final() && edge != s.final_transition_get())
      {
	set_state_final_angle(graph, s);
	mask = compute_cardinal_mask(mask,
				     GeometryProperties.getTransitionAngle
				     (s.final_transition_get()),
				     ANGLE_TOLERANCE);
      }

      // Return appropriate cardinal, if free.
      if ((mask & VJToolKitConstants.CARDINAL_N) == 0)
	return 90;
      if ((mask & VJToolKitConstants.CARDINAL_E) == 0)
	return 0;
      if ((mask & VJToolKitConstants.CARDINAL_S) == 0)
	return 270;
      if ((mask & VJToolKitConstants.CARDINAL_W) == 0)
	return 180;

      if ((mask & VJToolKitConstants.CARDINAL_NE) == 0)
	return 45;
      if ((mask & VJToolKitConstants.CARDINAL_SE) == 0)
	return 270 + 45;
      if ((mask & VJToolKitConstants.CARDINAL_SW) == 0)
	return 225;
      if ((mask & VJToolKitConstants.CARDINAL_NW) == 0)
	return 135;
      // No free cardinal was found.
      return VJToolKitConstants.NO_CARDINAL;
    }

    // Use here DefaultEdge since edge can be either Transition,
    // VcsnInitialTransition or VcsnFinalTransition.
    public static final int		get_free_angle(VJGraph graph,
						       State state,
						       DefaultEdge edge)
    {
      int res;

      if ((res = get_free_cardinal(graph, state, edge)) ==
	  VJToolKitConstants.NO_CARDINAL)
      {
	Debug.echo("Implement GeometricUtilities:get_free_angle()");
	return 0;
      }
      else
	return res;
    }


    public static final double		get_free_angle(VJGraph graph,
						       State state)
    {
      double res;

      if ((res = get_free_cardinal(graph, state, null)) ==
	  VJToolKitConstants.NO_CARDINAL)
      {
	Debug.echo("Implement GeometricUtilities:get_free_angle()");
	return 0;
      }
      else
	return res;
    }


    public static final int		compute_angle(Point2D p1,
						      Point2D p2)
    {
      // Get coordinates of points.
      double x1 = p1.getX();
      double y1 = p1.getY();
      double x2 = p2.getX();
      double y2 = p2.getY();

      // Normalize in p1.
      x2 -= x1;
      y2 -= y1;

      // Compute square length.
      double u2 = Math.sqrt(x2 * x2 + y2 *y2);

      // Compute angle.
      if (y2 > 0)
	return (int)Math.floor((2 * Math.PI - Math.acos(x2 / u2)) *
			       180 / Math.PI);
      else
	return (int)Math.floor((Math.acos(x2 / u2)) * 180 / Math.PI);
    }


    public static boolean		is_between_tolerance(double val,
							     double ref,
							     double tolerance)
    {
      return (ref - tolerance <= val) && (val <= ref + tolerance);
    }


    public static int			compute_cardinal_mask(int mask,
							      int alpha,
							      double tolerance)
    {
      if (is_between_tolerance((double)alpha, 0, tolerance))
	mask |= VJToolKitConstants.CARDINAL_E;
      else if (is_between_tolerance((double)alpha, 90, tolerance))
	mask |= VJToolKitConstants.CARDINAL_N;
      else if (is_between_tolerance((double)alpha, 180, tolerance))
	mask |= VJToolKitConstants.CARDINAL_W;
      else if (is_between_tolerance((double)alpha, 270, tolerance))
	mask |= VJToolKitConstants.CARDINAL_S;

      else if (is_between_tolerance((double)alpha, 45, tolerance))
	mask |= VJToolKitConstants.CARDINAL_NE;
      else if (is_between_tolerance((double)alpha, 135, tolerance))
	mask |= VJToolKitConstants.CARDINAL_NW;
      else if (is_between_tolerance((double)alpha, 225, tolerance))
	mask |= VJToolKitConstants.CARDINAL_SW;
      else if (is_between_tolerance(alpha, 315, tolerance))
	mask |= VJToolKitConstants.CARDINAL_SE;
      return mask;
    }

    /// FIXME: Recode this function.
    public static int		compute_angle(Point2D p1,
					      Point2D p2,
					      Point2D p3)
    {
      double x1 = p1.getX();
      double y1 = p1.getY();
      double x2 = p2.getX();
      double y2 = p2.getY();
      double x3 = p3.getX();
      double y3 = p3.getY();

      double ac_2 = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
      double ac = Math.sqrt(ac_2);
      double ab_2 = (x3 - x2) * (x3 - x2) + (y3 - y2) * (y3 - y2);
      double ab = Math.sqrt(ab_2);
      double bc_2 = (x3 - x1) * (x3 - x1) + (y3 - y1) * (y3 - y1);

      /// FIXME: Handle an exception here ?
      if (ab == 0 || ac == 0)
	return 0;
      return (int)Math.floor((Math.acos((ac_2 + ab_2 - bc_2) / ( 2 * ab * ac))) *
			     180 / Math.PI);
    }


    public static void		rotate(Point2D orig, Point2D p, double alpha)
    {
      double radius = Math.sqrt((p.getX() - orig.getX()) *
				(p.getX() - orig.getX()) + (p.getY() - orig.getY()) *
				(p.getY() - orig.getY()));

      double angle = ((double) compute_angle(orig, p)) * Math.PI / 180;
      p.setLocation(Math.cos(alpha + angle) * (radius) +
		    orig.getX(),
		    Math.sin(alpha + angle) * (radius) +
		    orig.getY());
    }


    public static void				set_state_initial_angle
    (VJGraph graph, State s)
    {
      if (s.is_initial())
      {
	GraphLayoutCache cache = graph.getGraphLayoutCache();
	CellView[] views =
	  cache.getMapping(new Object[]{s.initial_transition_get()});
	EdgeView edge = (EdgeView) views[0];

	if (edge != null)
	{
	  List points = edge.getPoints();
	  Point2D p1 = null;
	  Object p = points.get(1);
	  if (p instanceof PortView)
	    p1 = (Point2D) ((PortView) p).getLocation(null);
	  else
	    p1 = (Point2D) p;
	  Point2D p0 = (Point2D) points.get(0);

	  GeometryProperties.
	    setTransitionAngle(s.initial_transition_get(),
			       GeometricUtilities.compute_angle(p1, p0));
	}
      }
    }


    public static void			set_state_final_angle(VJGraph graph,
							      State s)
    {
      if (s.is_final())
      {
	GraphLayoutCache cache = graph.getGraphLayoutCache();
	CellView[] views =
	  cache.getMapping(new Object[]{s.final_transition_get()});
	EdgeView edge = (EdgeView) views[0];

	if (edge != null)
	{
	  List points = edge.getPoints();
	  Point2D p1 = null;
	  Object p = points.get(1);
	  if (p instanceof PortView)
	    p1 = (Point2D) ((PortView) p).getLocation(null);
	  else
	    p1 = (Point2D) p;
	  Point2D p0 = (Point2D) points.get(0);

	  GeometryProperties.
	    setTransitionAngle(s.final_transition_get(),
			       GeometricUtilities.compute_angle(p1,p0));
	}
      }
    }


    public static Point2D		get_Point2D(Object obj)
    {
      if (obj instanceof Point2D)
	return (Point2D) obj;
      if (obj instanceof PortView)
	return (Point2D) ((PortView) obj).getLocation(null);
      // Add here other possible cases.
      return new Point2D.Double();
    }

    public static Point2D		get_point_on_normal(Point2D from,
							    Point2D to,
							    double d)
    {
      double dy = from.getY() - to.getY();
      double dx = from.getX() - to.getX();

      // calculate mid-point of the main edge
      double midX = Math.min(from.getX(), to.getX())
	+ Math.abs(dx / 2);
      double midY = Math.min(from.getY(), to.getY())
	+ Math.abs(dy / 2);


      // modify the location of the control point along the
      // axis of the normal using the edge position
      double r = Math.sqrt(dx * dx + dy * dy);

      // convert polar coordinates to cartesian and
      // translate axis to the mid-point
      Point2D controlPoint = new Point2D.Double();
      controlPoint.setLocation(midX - d * (dy / r),
			       midY + d * (dx / r));

      return controlPoint;
    }

    // From top left to screen center
    public static Point2D		toOrigin(VJGraph g, Point2D p)
    {
	Point2D o = g.get_origin();
	Point2D translated = new Point2D.
	    Double((p.getX() - o.getX()) / VJToolKitConstants.PIXEL_PER_CM,
		   (p.getY() - o.getY()) / VJToolKitConstants.PIXEL_PER_CM);
	translated.setLocation(translated.getX(), -translated.getY());
	return translated;
    }

    // Convert a point which origin is screen center to top left origin
    public static Point2D		toScreen(VJGraph g, Point2D p)
    {
	Point2D o = g.get_origin();
	Point2D tp = new Point2D.Double(p.getX(), -p.getY());
	Point2D translated = new Point2D.
	    Double(o.getX() + tp.getX() * VJToolKitConstants.PIXEL_PER_CM,
		   o.getY() + tp.getY() * VJToolKitConstants.PIXEL_PER_CM);
	return translated;
    }

}
